/*
  MipSmooth  (c) 2003 by Klaus Post

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	The author can be contacted at:
	sh0dan@stofanet.dk
*/



void MipSmooth::isse_accumulate_line_org(const BYTE* c_plane, const BYTE** planeP, int planes, int rowsize, int t, int div) {
  __declspec(align(8)) static __int64 full = 0xffffffffffffffffi64;
  __declspec(align(8)) static __int64 add64 = (__int64)(16384) | ((__int64)(16384)<<32);
  __declspec(align(8)) __int64 t2 = t | ((__int64)t<<32);
  __declspec(align(8)) __int64 div64 = (__int64)(div) | ((__int64)(div)<<16) | ((__int64)(div)<<32) | ((__int64)(div)<<48);

  __asm {
    mov esi,c_plane;
    xor eax,eax          // EAX will be plane offset (all planes).
    align 16
testplane:
    mov ecx,[rowsize]
    cmp ecx, eax
    jle outloop

    movq mm0,[esi+eax]  // Load current frame pixels
    pxor mm6,mm6
     pxor mm7,mm7        
    movq mm4, mm0

    mov ecx,[planes]   // How many planes (this will be our counter)
    mov edi,[planeP];  // Adress of planeP array is now in edi
    lea edi,[edi+ecx*4]
    align 16
kernel_loop:
    mov edx,[edi]
    movq mm1,[edx+eax]      // Load 8 pixels from test plane
     movq mm2,mm0
    movq mm5, mm1           // Save test plane pixels (twice for unpack)
     pxor mm4,mm4
    pmaxub mm2,mm1          // Calc abs difference
     pminub mm1,mm0
    psubusb mm2,mm1         // mm2 = ads difference (packed bytes)
     psubusb mm2,[t2]         // Subtrack threshold (unsigned, so all below threshold will give 0)
    psubusb mm2,[t2]         // Subtrack threshold (unsigned, so all below threshold will give 0)
     movq mm1,mm5
    pcmpeqb mm2,mm4         // Compare values to 0
     prefetchnta [edx+eax+64]  // it might just help - and we have an idle CPU here anyway ;)
    movq mm3,mm2
     pxor mm2,[full]       // mm2 inverse mask
    movq mm4, mm0
     pand mm5, mm3
    pand mm4,mm2
     pxor mm1,mm1
    por mm4,mm5
    movq mm5,mm4  // stall (this & below)
    punpcklbw mm4,mm1         // mm4 = lower pixels
     punpckhbw mm5,mm1        // mm5 = upper pixels
    paddusw mm6,mm4
     paddusw mm7,mm5

    sub edi,4
    dec ecx
    jnz kernel_loop
     // Multiply (or in reality divides) added values, repack and store.
    movq mm4,[add64]
    pxor mm5,mm5
     movq mm0,mm6
    movq mm1,mm6
     punpcklwd mm0,mm5         // low,low
    movq mm6,[div64]
    punpckhwd mm1,mm5         // low,high
     movq mm2,mm7
    pmaddwd mm0,mm6            // pmaddwd is used due to it's better rounding.
     punpcklwd mm2,mm5         // high,low
     movq mm3,mm7
     paddd mm0,mm4
    pmaddwd mm1,mm6
     punpckhwd mm3,mm5         // high,high
     psrld mm0,15
     paddd mm1,mm4
    pmaddwd mm2,mm6
     packssdw mm0, mm0
     psrld mm1,15
     paddd mm2,mm4
    pmaddwd mm3,mm6
     packssdw mm1, mm1
     psrld mm2,15
     paddd mm3,mm4
    psrld mm3,15
     packssdw mm2, mm2
    packssdw mm3, mm3
     packuswb mm0,mm5
    packuswb mm1,mm5
     packuswb mm2,mm5
    packuswb mm3,mm5
     pshufw mm0,mm0,11111100b
    pshufw mm1,mm1,11110011b
     pshufw mm2,mm2,11001111b
    pshufw mm3,mm3,00111111b
     por mm0,mm1
    por mm2,mm3
    por mm0,mm2
    movq [esi+eax],mm0

    add eax,8   // Next 8 pixels
    jmp testplane
outloop:
    emms
  }
}

/**  Unrolled three plane weighed blend 
  *  Result is multiplied by 4+2+1 (org, first, second & third plane)  = 3+1 extra **/

void MipSmooth::isse_accumulate_line_org_weigh(const BYTE* c_plane, const BYTE** planeP,  int rowsize, int t, int div) {
  __declspec(align(8)) static __int64 full = 0xffffffffffffffffi64;
  __declspec(align(8)) static __int64 add64 = (__int64)(16384) | ((__int64)(16384)<<32);
  __declspec(align(8)) __int64 t2 = t | ((__int64)t<<32);
  __declspec(align(8)) __int64 div64 = (__int64)(div) | ((__int64)(div)<<16) | ((__int64)(div)<<32) | ((__int64)(div)<<48);

  __asm {
    mov esi,c_plane;
    xor eax,eax          // EAX will be plane offset (all planes).
    align 16
testplane:
    mov ecx,[rowsize]
    cmp ecx, eax
    jle outloop

    movq mm0,[esi+eax]  // Load current frame pixels
    pxor mm6,mm6
     pxor mm7,mm7        
    movq mm4, mm0

    mov edi,[planeP];  // Adress of planeP array is now in edi
    lea edi,[edi+2*4]

// LOOP 1 (Don't multiply)
    mov edx,[edi]
    movq mm1,[edx+eax]      // Load 8 pixels from test plane
     movq mm2,mm0
    movq mm5, mm1           // Save test plane pixels (twice for unpack)
     pxor mm4,mm4
    pmaxub mm2,mm1          // Calc abs difference
     pminub mm1,mm0
    psubusb mm2,mm1         // mm2 = ads difference (packed bytes)
     psubusb mm2,[t2]         // Subtrack threshold (unsigned, so all below threshold will give 0)
    psubusb mm2,[t2]         // Subtrack threshold (unsigned, so all below threshold will give 0)
     movq mm1,mm5
    pcmpeqb mm2,mm4         // Compare values to 0
     prefetchnta [edx+eax+64]  // it might just help - and we have an idle CPU here anyway ;)
    movq mm3,mm2
     pxor mm2,[full]       // mm2 inverse mask
    movq mm4, mm0
     pand mm5, mm3
    pand mm4,mm2
     pxor mm1,mm1
    por mm4,mm5
    movq mm5,mm4  // stall (this & below)
    punpcklbw mm4,mm1         // mm4 = lower pixels
     punpckhbw mm5,mm1        // mm5 = upper pixels
    paddusw mm6,mm4
     paddusw mm7,mm5

// LOOP 2 (Multiply by 2)
    mov edx,[edi-4]
    movq mm1,[edx+eax]      // Load 8 pixels from test plane
     movq mm2,mm0
    movq mm5, mm1           // Save test plane pixels (twice for unpack)
     pxor mm4,mm4
    pmaxub mm2,mm1          // Calc abs difference
     pminub mm1,mm0
    psubusb mm2,mm1         // mm2 = ads difference (packed bytes)
     psubusb mm2,[t2]         // Subtrack threshold (unsigned, so all below threshold will give 0)
    psubusb mm2,[t2]         // Subtrack threshold (unsigned, so all below threshold will give 0)
     movq mm1,mm5
    pcmpeqb mm2,mm4         // Compare values to 0
     prefetchnta [edx+eax+64]  // it might just help - and we have an idle CPU here anyway ;)
    movq mm3,mm2
     pxor mm2,[full]       // mm2 inverse mask
    movq mm4, mm0
     pand mm5, mm3
    pand mm4,mm2
     pxor mm1,mm1
    por mm4,mm5
    movq mm5,mm4  // stall (this & below)
    punpcklbw mm4,mm1         // mm4 = lower pixels
     punpckhbw mm5,mm1        // mm5 = upper pixels
    psllw mm4,1
     psllw mm5,1
    paddusw mm6,mm4
     paddusw mm7,mm5

// LOOP 2 (Multiply by 4)
    mov edx,[edi-8]
    movq mm1,[edx+eax]      // Load 8 pixels from test plane
     movq mm2,mm0
    movq mm5, mm1           // Save test plane pixels (twice for unpack)
     pxor mm4,mm4
    pmaxub mm2,mm1          // Calc abs difference
     pminub mm1,mm0
    psubusb mm2,mm1         // mm2 = ads difference (packed bytes)
     psubusb mm2,[t2]         // Subtrack threshold (unsigned, so all below threshold will give 0)
    psubusb mm2,[t2]         // Subtrack threshold (unsigned, so all below threshold will give 0)
     movq mm1,mm5
    pcmpeqb mm2,mm4         // Compare values to 0
     prefetchnta [edx+eax+64]  // it might just help - and we have an idle CPU here anyway ;)
    movq mm3,mm2
     pxor mm2,[full]       // mm2 inverse mask
    movq mm4, mm0
     pand mm5, mm3
    pand mm4,mm2
     pxor mm1,mm1
    por mm4,mm5
    movq mm5,mm4  // stall (this & below)
    punpcklbw mm4,mm1         // mm4 = lower pixels
     punpckhbw mm5,mm1        // mm5 = upper pixels
    psllw mm4,2
     psllw mm5,2
    paddusw mm6,mm4
     paddusw mm7,mm5


    // Multiply (or in reality divides) added values, repack and store.
    movq mm4,[add64]
    pxor mm5,mm5
     movq mm0,mm6
    movq mm1,mm6
     punpcklwd mm0,mm5         // low,low
    movq mm6,[div64]
    punpckhwd mm1,mm5         // low,high
     movq mm2,mm7
    pmaddwd mm0,mm6            // pmaddwd is used due to it's better rounding.
     punpcklwd mm2,mm5         // high,low
     movq mm3,mm7
     paddd mm0,mm4
    pmaddwd mm1,mm6
     punpckhwd mm3,mm5         // high,high
     psrld mm0,15
     paddd mm1,mm4
    pmaddwd mm2,mm6
     packssdw mm0, mm0
     psrld mm1,15
     paddd mm2,mm4
    pmaddwd mm3,mm6
     packssdw mm1, mm1
     psrld mm2,15
     paddd mm3,mm4
    psrld mm3,15
     packssdw mm2, mm2
    packssdw mm3, mm3
     packuswb mm0,mm5
    packuswb mm1,mm5
     packuswb mm2,mm5
    packuswb mm3,mm5
     pshufw mm0,mm0,11111100b
    pshufw mm1,mm1,11110011b
     pshufw mm2,mm2,11001111b
    pshufw mm3,mm3,00111111b
     por mm0,mm1
    por mm2,mm3
    por mm0,mm2
    movq [esi+eax],mm0

    add eax,8   // Next 8 pixels
    jmp testplane
outloop:
    emms
  }
}

/**  Unrolled three plane weighed blend SOFT
  *  Result is multiplied by 4+4+2+1 (org, first, second & third plane)  = **/

void MipSmooth::isse_accumulate_line_org_soft_weigh(const BYTE* c_plane, const BYTE** planeP,  int rowsize, int t, int div) {
  __declspec(align(8)) static __int64 full = 0xffffffffffffffffi64;
  __declspec(align(8)) static __int64 add64 = (__int64)(16384) | ((__int64)(16384)<<32);
  __declspec(align(8)) __int64 t2 = t | ((__int64)t<<32);
  __declspec(align(8)) __int64 div64 = (__int64)(div) | ((__int64)(div)<<16) | ((__int64)(div)<<32) | ((__int64)(div)<<48);


  __asm {
    mov esi,c_plane;
    xor eax,eax          // EAX will be plane offset (all planes).
    align 16
testplane:
    mov ecx,[rowsize]
    cmp ecx, eax
    jle outloop

    movq mm0,[esi+eax]  // Load current frame pixels
      pxor mm2,mm2
    movq mm6,mm0
     movq mm7,mm0
    punpcklbw mm6,mm2    // mm0 = lower 4 pixels  (exhanging h/l in these two give funny results)
     punpckhbw mm7,mm2     // mm1 = upper 4 pixels
    psllw mm6,2
     psllw mm7,2
    movq mm4, mm0

    mov edi,[planeP];  // Adress of planeP array is now in edi
    lea edi,[edi+2*4]

// LOOP 1 (Don't multiply)
    mov edx,[edi]
    movq mm1,[edx+eax]      // Load 8 pixels from test plane
     movq mm2,mm0
    movq mm5, mm1           // Save test plane pixels (twice for unpack)
     pxor mm4,mm4
    pmaxub mm2,mm1          // Calc abs difference
     pminub mm1,mm0
    psubusb mm2,mm1         // mm2 = ads difference (packed bytes)
     psubusb mm2,[t2]         // Subtrack threshold (unsigned, so all below threshold will give 0)
    psubusb mm2,[t2]         // Subtrack threshold (unsigned, so all below threshold will give 0)
     movq mm1,mm5
    pcmpeqb mm2,mm4         // Compare values to 0
     prefetchnta [edx+eax+64]  // it might just help - and we have an idle CPU here anyway ;)
    movq mm3,mm2
     pxor mm2,[full]       // mm2 inverse mask
    movq mm4, mm0
     pand mm5, mm3
    pand mm4,mm2
     pxor mm1,mm1
    por mm4,mm5
    movq mm5,mm4  // stall (this & below)
    punpcklbw mm4,mm1         // mm4 = lower pixels
     punpckhbw mm5,mm1        // mm5 = upper pixels
    paddusw mm6,mm4
     paddusw mm7,mm5

// LOOP 2 (Multiply by 2)
    mov edx,[edi-4]
    movq mm1,[edx+eax]      // Load 8 pixels from test plane
     movq mm2,mm0
    movq mm5, mm1           // Save test plane pixels (twice for unpack)
     pxor mm4,mm4
    pmaxub mm2,mm1          // Calc abs difference
     pminub mm1,mm0
    psubusb mm2,mm1         // mm2 = ads difference (packed bytes)
     psubusb mm2,[t2]         // Subtrack threshold (unsigned, so all below threshold will give 0)
    psubusb mm2,[t2]         // Subtrack threshold (unsigned, so all below threshold will give 0)
     movq mm1,mm5
    pcmpeqb mm2,mm4         // Compare values to 0
     prefetchnta [edx+eax+64]  // it might just help - and we have an idle CPU here anyway ;)
    movq mm3,mm2
     pxor mm2,[full]       // mm2 inverse mask
    movq mm4, mm0
     pand mm5, mm3
    pand mm4,mm2
     pxor mm1,mm1
    por mm4,mm5
    movq mm5,mm4  // stall (this & below)
    punpcklbw mm4,mm1         // mm4 = lower pixels
     punpckhbw mm5,mm1        // mm5 = upper pixels
    psllw mm4,1
     psllw mm5,1
    paddusw mm6,mm4
     paddusw mm7,mm5

// LOOP 2 (Multiply by 4)
    mov edx,[edi-8]
    movq mm1,[edx+eax]      // Load 8 pixels from test plane
     movq mm2,mm0
    movq mm5, mm1           // Save test plane pixels (twice for unpack)
     pxor mm4,mm4
    pmaxub mm2,mm1          // Calc abs difference
     pminub mm1,mm0
    psubusb mm2,mm1         // mm2 = ads difference (packed bytes)
     psubusb mm2,[t2]         // Subtrack threshold (unsigned, so all below threshold will give 0)
    psubusb mm2,[t2]         // Subtrack threshold (unsigned, so all below threshold will give 0)
     movq mm1,mm5
    pcmpeqb mm2,mm4         // Compare values to 0
     prefetchnta [edx+eax+64]  // it might just help - and we have an idle CPU here anyway ;)
    movq mm3,mm2
     pxor mm2,[full]       // mm2 inverse mask
    movq mm4, mm0
     pand mm5, mm3
    pand mm4,mm2
     pxor mm1,mm1
    por mm4,mm5
    movq mm5,mm4  // stall (this & below)
    punpcklbw mm4,mm1         // mm4 = lower pixels
     punpckhbw mm5,mm1        // mm5 = upper pixels
    psllw mm4,2
     psllw mm5,2
    paddusw mm6,mm4
     paddusw mm7,mm5


    // Multiply (or in reality divides) added values, repack and store.
    movq mm4,[add64]
    pxor mm5,mm5
     movq mm0,mm6
    movq mm1,mm6
     punpcklwd mm0,mm5         // low,low
    movq mm6,[div64]
    punpckhwd mm1,mm5         // low,high
     movq mm2,mm7
    pmaddwd mm0,mm6            // pmaddwd is used due to it's better rounding.
     punpcklwd mm2,mm5         // high,low
     movq mm3,mm7
     paddd mm0,mm4
    pmaddwd mm1,mm6
     punpckhwd mm3,mm5         // high,high
     psrld mm0,15
     paddd mm1,mm4
    pmaddwd mm2,mm6
     packssdw mm0, mm0
     psrld mm1,15
     paddd mm2,mm4
    pmaddwd mm3,mm6
     packssdw mm1, mm1
     psrld mm2,15
     paddd mm3,mm4
    psrld mm3,15
     packssdw mm2, mm2
    packssdw mm3, mm3
     packuswb mm0,mm5
    packuswb mm1,mm5
     packuswb mm2,mm5
    packuswb mm3,mm5
     pshufw mm0,mm0,11111100b
    pshufw mm1,mm1,11110011b
     pshufw mm2,mm2,11001111b
    pshufw mm3,mm3,00111111b
     por mm0,mm1
    por mm2,mm3
    por mm0,mm2
    movq [esi+eax],mm0

    add eax,8   // Next 8 pixels
    jmp testplane
outloop:
    emms
  }
}



void MipSmooth::isse_accumulate_line_org_soft(const BYTE* c_plane, const BYTE** planeP, int planes, int rowsize, int t, int div) {
  __declspec(align(8)) static __int64 full = 0xffffffffffffffffi64;
  __declspec(align(8)) static __int64 add64 = (__int64)(16384) | ((__int64)(16384)<<32);
  __declspec(align(8)) __int64 t2 = t | ((__int64)t<<32);
  __declspec(align(8)) __int64 div64 = (__int64)(div) | ((__int64)(div)<<16) | ((__int64)(div)<<32) | ((__int64)(div)<<48);

  __asm {
    mov esi,c_plane;
    xor eax,eax          // EAX will be plane offset (all planes).
    align 16
testplane:
    mov ecx,[rowsize]
    cmp ecx, eax
    jle outloop

    movq mm0,[esi+eax]  // Load current frame pixels
    pxor mm2,mm2
    movq mm6,mm0
     movq mm7,mm0
    punpcklbw mm6,mm2    // mm0 = lower 4 pixels  (exhanging h/l in these two give funny results)
     punpckhbw mm7,mm2     // mm1 = upper 4 pixels
    movq mm4, mm0

    mov ecx,[planes]   // How many planes (this will be our counter)
    mov edi,[planeP];  // Adress of planeP array is now in edi
    lea edi,[edi+ecx*4]
    align 16
kernel_loop:
    mov edx,[edi]
    movq mm1,[edx+eax]      // Load 8 pixels from test plane
     movq mm2,mm0
    movq mm5, mm1           // Save test plane pixels (twice for unpack)
     pxor mm4,mm4
    pmaxub mm2,mm1          // Calc abs difference
     pminub mm1,mm0
    psubusb mm2,mm1         // mm2 = ads difference (packed bytes)
     psubusb mm2,[t2]         // Subtrack threshold (unsigned, so all below threshold will give 0)
    psubusb mm2,[t2]         // Subtrack threshold (unsigned, so all below threshold will give 0)
     movq mm1,mm5
    pcmpeqb mm2,mm4         // Compare values to 0
     prefetchnta [edx+eax+64]  // it might just help - and we have an idle CPU here anyway ;)
    movq mm3,mm2
     pxor mm2,[full]       // mm2 inverse mask
    movq mm4, mm0
     pand mm5, mm3
    pand mm4,mm2
     pxor mm1,mm1
    por mm4,mm5
    movq mm5,mm4  // stall (this & below)
    punpcklbw mm4,mm1         // mm4 = lower pixels
     punpckhbw mm5,mm1        // mm5 = upper pixels
    paddusw mm6,mm4
     paddusw mm7,mm5

    sub edi,4
    dec ecx
    jnz kernel_loop
     // Multiply (or in reality divides) added values, repack and store.
    movq mm4,[add64]
    pxor mm5,mm5
     movq mm0,mm6
    movq mm1,mm6
     punpcklwd mm0,mm5         // low,low
    movq mm6,[div64]
    punpckhwd mm1,mm5         // low,high
     movq mm2,mm7
    pmaddwd mm0,mm6            // pmaddwd is used due to it's better rounding.
     punpcklwd mm2,mm5         // high,low
     movq mm3,mm7
     paddd mm0,mm4
    pmaddwd mm1,mm6
     punpckhwd mm3,mm5         // high,high
     psrld mm0,15
     paddd mm1,mm4
    pmaddwd mm2,mm6
     packssdw mm0, mm0
     psrld mm1,15
     paddd mm2,mm4
    pmaddwd mm3,mm6
     packssdw mm1, mm1
     psrld mm2,15
     paddd mm3,mm4
    psrld mm3,15
     packssdw mm2, mm2
    packssdw mm3, mm3
     packuswb mm0,mm5
    packuswb mm1,mm5
     packuswb mm2,mm5
    packuswb mm3,mm5
     pshufw mm0,mm0,11111100b
    pshufw mm1,mm1,11110011b
     pshufw mm2,mm2,11001111b
    pshufw mm3,mm3,00111111b
     por mm0,mm1
    por mm2,mm3
    por mm0,mm2
    movq [esi+eax],mm0

    add eax,8   // Next 8 pixels
    jmp testplane
outloop:
    emms
  }
}

/**
 * This function is very close to the temporalsoften mode 2 filter.
 *
 * The following pseudocode describes how it is done:
 *
 * For all pixels:
 *
 * int c_pixel = (mode_soft) ? src[x] : 0;
 * int last_pixel = src[x];
 *
 * for (i=0;i<planes;i++) 
 *   if (abs(mip_plane[x]-src[x]) > threshold) {
 *     c_pixel +=  last_pixel;
 *   } else {
 *     c_pixel += mip_plane[x];
 *     last_pixel = mip_plane[x];
 *   }
 * }
 * int div = (mode_soft) planes : planes-1;
 * dst[x] = c_pixel / div;
 *
 * (c) Copyright 2003 Klaus Post
 *
 **/


void MipSmooth::isse_accumulate_line_strong(const BYTE* c_plane, const BYTE** planeP, int planes, int rowsize, BYTE* t, int div) {
  __int64 full = 0xffffffffffffffffi64;
  __int64 add64 = (__int64)(16384) | ((__int64)(16384)<<32);

  __int64 div64 = (__int64)(div) | ((__int64)(div)<<16) | ((__int64)(div)<<32) | ((__int64)(div)<<48);

  __asm {
    mov esi,c_plane;
    xor eax,eax          // EAX will be plane offset (all planes).
    align 16
testplane:
    mov ebx,[rowsize]
    cmp ebx, eax
    jle outloop

    movq mm0,[esi+eax]  // Load current frame pixels
    pxor mm6,mm6
     pxor mm7,mm7        
    movq mm4, mm0

    mov ebx,[planes]   // How many planes (this will be our counter)
    mov ecx, [t]
    mov edi,[planeP];  // Adress of planeP array is now in edi
    lea ecx,[ecx+ebx*8]
    lea edi,[edi+ebx*4]
    align 16
kernel_loop:
    mov edx,[edi]
    movq mm1,[edx+eax]      // Load 8 pixels from test plane
     movq mm2,mm0
    movq mm5, mm1           // Save test plane pixels (twice for unpack)
     pxor mm3,mm3
    pmaxub mm2,mm1          // Calc abs difference
     pminub mm1,mm0
    psubusb mm2,mm1         // mm2 = ads difference (packed bytes)
    psubusb mm2,[ecx]         // Subtrack threshold (unsigned, so all below threshold will give 0)
     movq mm1,mm5
    pcmpeqb mm2,mm3         // Compare values to 0
     prefetchnta [edx+eax+64]  // it might just help - and we have an idle CPU here anyway ;)
    movq mm3,mm2
     pxor mm2,[full]          // mm2 inverse mask
     pand mm5, mm3
    pand mm4,mm2
     pxor mm1,mm1
    por mm4,mm5
    movq mm3,mm4                // Overwrite previous 
    movq mm5,mm4              // stall (this & below)
    punpcklbw mm3,mm1         // mm4 = lower pixels
     punpckhbw mm5,mm1        // mm5 = upper pixels
    paddusw mm6,mm3
     paddusw mm7,mm5

    sub edi,4
    sub ecx,8
    dec ebx
    jnz kernel_loop
     // Multiply (or in reality divides) added values, repack and store.
    movq mm4,[add64]
    pxor mm5,mm5
     movq mm0,mm6
    movq mm1,mm6
     punpcklwd mm0,mm5         // low,low
    movq mm6,[div64]
    punpckhwd mm1,mm5         // low,high
     movq mm2,mm7
    pmaddwd mm0,mm6            // pmaddwd is used due to it's better rounding.
     punpcklwd mm2,mm5         // high,low
     movq mm3,mm7
     paddd mm0,mm4
    pmaddwd mm1,mm6
     punpckhwd mm3,mm5         // high,high
     psrld mm0,15
     paddd mm1,mm4
    pmaddwd mm2,mm6
     packssdw mm0, mm0
     psrld mm1,15
     paddd mm2,mm4
    pmaddwd mm3,mm6
     packssdw mm1, mm1
     psrld mm2,15
     paddd mm3,mm4
    psrld mm3,15
     packssdw mm2, mm2
    packssdw mm3, mm3
     packuswb mm0,mm5
    packuswb mm1,mm5
     packuswb mm2,mm5
    packuswb mm3,mm5
     pshufw mm0,mm0,11111100b
    pshufw mm1,mm1,11110011b
     pshufw mm2,mm2,11001111b
    pshufw mm3,mm3,00111111b
     por mm0,mm1
    por mm2,mm3
    por mm0,mm2
    movq [esi+eax],mm0

    add eax,8   // Next 8 pixels
    jmp testplane
outloop:
    emms
  }
}

void MipSmooth::isse_accumulate_line_soft(const BYTE* c_plane, const BYTE** planeP, int planes, int rowsize, BYTE* t, int div) {
  __int64 full = 0xffffffffffffffffi64;
  __int64 add64 = (__int64)(16384) | ((__int64)(16384)<<32);

  __int64 div64 = (__int64)(div) | ((__int64)(div)<<16) | ((__int64)(div)<<32) | ((__int64)(div)<<48);

  __asm {
    mov esi,c_plane;
    xor eax,eax          // EAX will be plane offset (all planes).
    align 16
testplane:
    mov ebx,[rowsize]
    cmp ebx, eax
    jle outloop

    movq mm0,[esi+eax]  // Load current frame pixels
    pxor mm2,mm2
    movq mm6,mm0
     movq mm7,mm0
    punpcklbw mm6,mm2    // mm0 = lower 4 pixels  (exhanging h/l in these two give funny results)
     punpckhbw mm7,mm2     // mm1 = upper 4 pixels
    movq mm4, mm0

    mov ebx,[planes]   // How many planes (this will be our counter)
    mov ecx, [t]
    mov edi,[planeP];  // Adress of planeP array is now in edi
    lea ecx,[ecx+ebx*8]
    lea edi,[edi+ebx*4]
    align 16
kernel_loop:
    mov edx,[edi]
    movq mm1,[edx+eax]      // Load 8 pixels from test plane
     movq mm2,mm0
    movq mm5, mm1           // Save test plane pixels (twice for unpack)
     pxor mm3,mm3
    pmaxub mm2,mm1          // Calc abs difference
     pminub mm1,mm0
    psubusb mm2,mm1         // mm2 = ads difference (packed bytes)
    psubusb mm2,[ecx]         // Subtrack threshold (unsigned, so all below threshold will give 0)
     movq mm1,mm5
    pcmpeqb mm2,mm3         // Compare values to 0
     prefetchnta [edx+eax+64]  // it might just help - and we have an idle CPU here anyway ;)
    movq mm3,mm2
     pxor mm2,[full]          // mm2 inverse mask
     pand mm5, mm3
    pand mm4,mm2
     pxor mm1,mm1
    por mm4,mm5
    movq mm3,mm4  // TEST!!!
    movq mm5,mm4  // stall (this & below)
    punpcklbw mm3,mm1         // mm4 = lower pixels
     punpckhbw mm5,mm1        // mm5 = upper pixels
    paddusw mm6,mm3
     paddusw mm7,mm5

    sub edi,4
    sub ecx,8
    dec ebx
    jnz kernel_loop
     // Multiply (or in reality divides) added values, repack and store.
    movq mm4,[add64]
    pxor mm5,mm5
     movq mm0,mm6
    movq mm1,mm6
     punpcklwd mm0,mm5         // low,low
    movq mm6,[div64]
    punpckhwd mm1,mm5         // low,high
     movq mm2,mm7
    pmaddwd mm0,mm6            // pmaddwd is used due to it's better rounding.
     punpcklwd mm2,mm5         // high,low
     movq mm3,mm7
     paddd mm0,mm4
    pmaddwd mm1,mm6
     punpckhwd mm3,mm5         // high,high
     psrld mm0,15
     paddd mm1,mm4
    pmaddwd mm2,mm6
     packssdw mm0, mm0
     psrld mm1,15
     paddd mm2,mm4
    pmaddwd mm3,mm6
     packssdw mm1, mm1
     psrld mm2,15
     paddd mm3,mm4
    psrld mm3,15
     packssdw mm2, mm2
    packssdw mm3, mm3
     packuswb mm0,mm5
    packuswb mm1,mm5
     packuswb mm2,mm5
    packuswb mm3,mm5
     pshufw mm0,mm0,11111100b
    pshufw mm1,mm1,11110011b
     pshufw mm2,mm2,11001111b
    pshufw mm3,mm3,00111111b
     por mm0,mm1
    por mm2,mm3
    por mm0,mm2
    movq [esi+eax],mm0

    add eax,8   // Next 8 pixels
    jmp testplane
outloop:
    emms
  }
}

int MipSmooth::isse_scenechange(const BYTE* c_plane, const BYTE* tplane, int height, int width, int c_pitch, int t_pitch, int pixel_type) {
  int wp=(width/32)*32;
  int hp=height;
  int returnvalue=0xbadbad00;
  int mask=0xffffffff;
  if (pixel_type == VideoInfo::CS_BGR32)
    mask = 0x00ffffff;
  if (pixel_type == VideoInfo::CS_YUY2)
    mask = 0x00ff00ff;

  __asm {
    xor ebx,ebx     // Height
    movd mm5,[mask]  // Pixel mask
    pshufw mm5,mm5,01000100b  // Copy mask to upper
    mov edx, c_pitch    //copy pitch
    mov ecx, t_pitch    //copy pitch
    pxor mm6,mm6   // We maintain two sums, for better pairablility
    pxor mm7,mm7
    mov esi, c_plane
    mov edi, tplane
    jmp yloopover
    align 16
yloop:
    inc ebx
    add edi,ecx     // add pitch to both planes
    add esi,edx
yloopover:
    cmp ebx,[hp]
    jge endframe
    xor eax,eax     // Width
    align 16
xloop:
    cmp eax,[wp]    
    jge yloop

    movq mm0,[esi+eax]
     movq mm2,[esi+eax+8]
    pand mm0,mm5
    movq mm1,[edi+eax]
     pand mm2,mm5
    movq mm3,[edi+eax+8]
     pand mm1,mm5
    pand mm3,mm5
    psadbw mm0,mm1    // Sum of absolute difference
     psadbw mm2,mm3
    paddd mm6,mm0     // Add...
     paddd mm7,mm2
    movq mm0,[esi+eax+16]
     movq mm2,[esi+eax+24]
    pand mm0,mm5
     movq mm1,[edi+eax+16]
    pand mm2,mm5
     movq mm3,[edi+eax+24]
    pand mm1,mm5
     pand mm3,mm5
    psadbw mm0,mm1
     psadbw mm2,mm3
    paddd mm6,mm0
     paddd mm7,mm2

    add eax,32
    jmp xloop
endframe:
    paddd mm7,mm6
    movd returnvalue,mm7
    emms
  }
  return returnvalue;
}




void MipSmooth::isse_blendline_soft(const BYTE* c_plane, const BYTE** planeP, int planes, int rowsize, int div) {
  __declspec(align(8)) static __int64 full = 0xffffffffffffffffi64;
  __declspec(align(8)) static __int64 add64 = (__int64)(16384) | ((__int64)(16384)<<32);
//  __declspec(align(8)) __int64 t2 = t | ((__int64)t<<32);
  __declspec(align(8)) __int64 div64 = (__int64)(div) | ((__int64)(div)<<16) | ((__int64)(div)<<32) | ((__int64)(div)<<48);

  __asm {
    mov esi,c_plane;
    xor eax,eax          // EAX will be plane offset (all planes).
    align 16
testplane:
    mov ecx,[rowsize]
    cmp ecx, eax
    jle outloop

    movq mm0,[esi+eax]  // Load current frame pixels
    pxor mm2,mm2
    movq mm6,mm0
     movq mm7,mm0
    punpcklbw mm6,mm2    // mm0 = lower 4 pixels  (exhanging h/l in these two give funny results)
     punpckhbw mm7,mm2     // mm1 = upper 4 pixels
    movq mm4, mm0

    mov ecx,[planes]   // How many planes (this will be our counter)
    mov edi,[planeP];  // Adress of planeP array is now in edi
    lea edi,[edi+ecx*4]
    align 16
kernel_loop:
    mov edx,[edi]
    movq mm1,[edx+eax]      // Load 8 pixels from test plane
    movq mm5,mm1
    punpcklbw mm1,mm2         // mm4 = lower pixels
     punpckhbw mm5,mm2        // mm5 = upper pixels
    paddusw mm6,mm1
     paddusw mm7,mm5
    sub edi,4
    dec ecx
    jnz kernel_loop
     // Multiply (or in reality divides) added values, repack and store.
    movq mm4,[add64]
    pxor mm5,mm5
     movq mm0,mm6
    movq mm1,mm6
     punpcklwd mm0,mm5         // low,low
    movq mm6,[div64]
    punpckhwd mm1,mm5         // low,high
     movq mm2,mm7
    pmaddwd mm0,mm6            // pmaddwd is used due to it's better rounding.
     punpcklwd mm2,mm5         // high,low
     movq mm3,mm7
     paddd mm0,mm4
    pmaddwd mm1,mm6
     punpckhwd mm3,mm5         // high,high
     psrld mm0,15
     paddd mm1,mm4
    pmaddwd mm2,mm6
     packssdw mm0, mm0
     psrld mm1,15
     paddd mm2,mm4
    pmaddwd mm3,mm6
     packssdw mm1, mm1
     psrld mm2,15
     paddd mm3,mm4
    psrld mm3,15
     packssdw mm2, mm2
    packssdw mm3, mm3
     packuswb mm0,mm5
    packuswb mm1,mm5
     packuswb mm2,mm5
    packuswb mm3,mm5
     pshufw mm0,mm0,11111100b
    pshufw mm1,mm1,11110011b
     pshufw mm2,mm2,11001111b
    pshufw mm3,mm3,00111111b
     por mm0,mm1
    por mm2,mm3
    por mm0,mm2
    movq [esi+eax],mm0

    add eax,8   // Next 8 pixels
    jmp testplane
outloop:
    emms
  }
}
